#ifndef TLM_FIFO_H
#define TLM_FIFO_H

#include <scoot_tags.h>
#include <sc_event.h>
#include <sc_wait.h>

namespace tlm {

class tlm_fifo_base: public scoot_channel
{
    public:
    tlm_fifo_base(){}

    // Events checking

    const sc_core::sc_event& data_read_event() const
    {
        return _data_read;
    }
    const sc_core::sc_event& data_written_event() const
    {
        return _data_written;
    }

    protected:
    // Disabled
    tlm_fifo_base(const tlm_fifo_base&);

    sc_core::sc_event _data_read;
    sc_core::sc_event _data_written;
};

template <class T>
class tlm_fifo: public tlm_fifo_base
{
    public:

    // Constructors

    explicit tlm_fifo( int size = 16 ):
        capacity(size), offset(0), size_rd(0)
    {
        data = new T[size];
    }

    explicit tlm_fifo( const char* name, int size = 16):
        capacity(size), offset(0), size_rd(0)
        #ifdef SCOOT_SKIP_NOTIFICATION_PHASES
        , data_read(false), data_written(false)
        #endif
    {
        data = new T[size];
    }

    // Destructors

    ~tlm_fifo();

    // Get interface

    void get(T& t)
    {
        while(size_rd == offset)
        {
		sc_core::wait(data_written_event());
        }

        t = data[size_rd - 1];

        size_rd--;

        #ifdef SCOOT_SKIP_NOTIFICATION_PHASES
        notify_read = true;
        #else
        _data_read.notify(0);
        #endif
    }

    T get()
    {
	 while(size_rd == offset)
        {
		sc_core::wait(data_written_event());
        }

        T t = data[size_rd - 1];

        size_rd--;

        #ifdef SCOOT_SKIP_NOTIFICATION_PHASES
        notify_read = true;
        #else
        _data_read.notify(0);
        #endif
    
	return t;
    }

    bool nb_get( T& t)
    {
        if(size_rd == offset)
        {
            return false;
        }

        t = data[size_rd - 1];
        size_rd--;

        #ifdef SCOOT_SKIP_NOTIFICATION_PHASES
        notify_read = true;
        #else
        _data_read.notify(0);
	#endif

        return true;
    }

    bool nb_can_get() const
    {
	    return size_rd != offset;
    }

    // Put interface

    void put( const T& t)
    {
        while(size_rd == capacity)
        {
		sc_core::wait(data_read_event());
        }

        for(int i = size_rd; i >= 0; i--)
        {
           data[i] = data[i-1];
        }

        offset++;
        size_rd++;

        data[0] = t;

        #ifdef SCOOT_SKIP_NOTIFICATION_PHASES
        notify_written = true;
        #else
        _data_written.notify(0);
        #endif
    }

    bool nb_put( const T& t)
    {
        if(size_rd == capacity)
        {
            return false;
        }

        for(int i = size_rd; i >= 0; i--)
        {
           data[i] = data[i-1];
        }

        offset++;
        size_rd++;

        data[0] = t;

        #ifdef SCOOT_SKIP_NOTIFICATION_PHASES
        notify_written = true;
        #else
        _data_written.notify(0);
        #endif

        return true;
    }

    bool nb_can_put() const
    {
	    return size_rd != capacity;
    }

    // Peek interface
    
    void peek(T& t) const
    {
	while(size_rd == offset)
        {
		sc_core::wait(data_written_event());
        }

        t = data[size_rd - 1];
    }

    
    T peek() const
    {
	while(size_rd == offset)
        {
		sc_core::wait(data_written_event());
        }

        return data[size_rd - 1];
    }

    bool nb_peek(T& t)const
    {
	    if(size_rd == offset)
	    {
		    return false;
	    }

	    t =  data[size_rd - 1];
    }
	    
private:

    // Update

    void update()
    {
        offset = 0;

        #ifdef SCOOT_SKIP_NOTIFICATION_PHASES
        _data_read.do_notify(notify_read);
        _data_written.do_notify(notify_written);
        notify_read = false;
        notify_written = false;
        #endif
    }

    // Disabled
    tlm_fifo( const tlm_fifo<T>& );
    tlm_fifo& operator= ( const tlm_fifo<T>& );

    T* data;
    const int capacity;
    int offset;
    int size_rd;

    #ifdef SCOOT_SKIP_NOTIFICATION_PHASES
    bool notify_read;
    bool notify_written;
    #endif
};

}

#endif
